/**
 * \file        notifier.cpp
 * \author      Plugware Solutions.com, Ltd.
 * \date        
 * \brief       [brief here]
 *
 *              Copyright (c) 2003
 *              Plugware Solutions.com, Ltd.
 *
 *              Permission to use, copy, modify, distribute and sell 
 *              this software and its documentation for any purpose 
 *              is hereby granted without fee, provided that the above 
 *              copyright notice appear in all copies and the both that 
 *              copyright notice and this permission notice appear 
 *              in supporting documentation.  Plugware Solutions.com, Ltd. 
 *              makes no reresentations about the suitability of this 
 *              software for any purpose.  It is provided "as is" 
 *              without express or implied warranty.
 */

/*
**  Includes
*/

#define _WIN32_WINNT 0x0500
#include <math.h>
#include <float.h>
#include <conio.h>
#include <iostream>
#include "notifier.h"
using namespace plugware;

/*
**  Implemenation
*/

enum events
{
    spawn,
    die
};

struct input : core::work_unit
{
    void process();
    input(HANDLE h_event);
    HANDLE m_event;
};

struct work : core::work_unit, subscriber<events> 
{
    work();
    void process();
    void on_notify(const events& e);
    
    bool   m_work;
    double m_data;
};

///////////////////////////////////////////////////////////////////////
// main
///////////////////////////////////////////////////////////////////////
int main(int, char**)
{
    // initialize our thread pool
    global::thread_pool::instance().initialize();
    HANDLE h_event = ::CreateEvent(0, 0, 0, 0);
 
    // start work 
    global::thread_pool::instance().queue_request(new work);
    
    // NOTE: most of these pauses exist to keep the output
    //       nice and clean -- these are unnecessary if you don't care
    //       about the output.  The only important sleep is the one at
    //       at the end; the article describes why.
    
    // wait for it to come online 
    ::Sleep(1000);
    
    // and then kill it asynchronously
    global::thread_pool::instance().queue_request(new async_notifier<events>(die));
    
    // and wait for it to die
    ::Sleep(1000);
    
    // start input and work
    global::thread_pool::instance().queue_request(new input(h_event));
    global::thread_pool::instance().queue_request(new work);
    
    // Wait for death, then pause a moment to let our work die
    ::WaitForSingleObject(h_event, INFINITE);
    ::Sleep(5 * 1000);
	return 0;
}

///////////////////////////////////////////////////////////////////////
// work
///////////////////////////////////////////////////////////////////////
work::work() 
: m_work(true), m_data(0) 
{
    std::cout << "alive: " << this << std::endl;
}

void work::process() 
{
    // we subscribe in process because
    // we are created during notification.  if
    // we were to subscribe during work::work
    // we would be added to the notifier map
    // and spawning would never cease to end!
    subscribe();
    
    // simulate work and requeue if necessary
    m_data += DBL_EPSILON * ::sin(m_data);
    if (m_work) global::thread_pool::instance().queue_request(this);
}

void work::on_notify(const events& e) 
{
    switch (e) 
    {
        case die:   
            unsubscribe();
            std::cout << "dying: " << this << std::endl; 
            m_work = false; 
            break;
        
        case spawn: 
            std::cout << "spawing: " << this << std::endl;
            global::thread_pool::instance().queue_request(new work); 
            break;
            
        default: 
            _ASSERT(false);
    }
}

///////////////////////////////////////////////////////////////////////
// input
///////////////////////////////////////////////////////////////////////
input::input(HANDLE h_event) 
: m_event(h_event) 
{
}

void input::process()
{
    std::cout << "(S)pawn/(D)ie: ";
    switch (::tolower(::getch()))
    {
        // die and leave
        case 'd': 
            notifier<events>::instance().notify(die);
            ::SetEvent(m_event);
            return;
            
        case 's':
            notifier<events>::instance().notify(spawn);
            break;
    }

    global::thread_pool::instance().queue_request(this);
}